#
# Copyright (c) 2004-2010 Purdue University All rights reserved.
# 
# Developed by: HUBzero Technology Group, Purdue University
#               http://hubzero.org
# 
# HUBzero is free software: you can redistribute it and/or modify it under the terms of the
# GNU Lesser General Public License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
# 
# HUBzero is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details.  You should have received a
# copy of the GNU Lesser General Public License along with HUBzero.
# If not, see <http://www.gnu.org/licenses/>.
# 
# GNU LESSER GENERAL PUBLIC LICENSE
# Version 3, 29 June 2007
# Copyright (C) 2007 Free Software Foundation, Inc. <http://fsf.org/>
#
import os.path
import re

from TimeConversion import minTohhmmss

class RemoteBatchPBS:
   def __init__(self,
                localJobId,
                appScriptName,
                environment,
                transferExecutable,
                executable,
                arguments,
                stdinput,
                isMultiCoreRequest,
                computationMode,
                preManagerCommands,
                managerCommand,
                postManagerCommands,
                nNodes,
                ppn,
                hostAttributes,
                remoteBatchQueue,
                wallTime,
                remoteBatchAccount,
                quotaLimit,
                TIMESTAMPSTART,
                TIMESTAMPFINISH,
                TIMERESULTS):
      self.localJobId          = localJobId
      self.appScriptName       = appScriptName
      self.environment         = environment
      self.transferExecutable  = transferExecutable
      if transferExecutable:
         self.executable       = os.path.join(".",os.path.basename(executable))
      else:
         self.executable       = executable
      self.arguments           = arguments
      self.stdinput            = stdinput
      self.isMultiCoreRequest  = isMultiCoreRequest
      self.computationMode     = computationMode
      self.preManagerCommands  = preManagerCommands
      self.managerCommand      = managerCommand
      self.postManagerCommands = postManagerCommands
      self.nNodes              = nNodes
      self.ppn                 = ppn
      self.hostAttributes      = hostAttributes
      self.remoteBatchQueue    = remoteBatchQueue
      self.wallTime            = minTohhmmss(wallTime)
      if remoteBatchAccount != "":
         self.account          = '#PBS -A ' + remoteBatchAccount
      else:
         self.account          = ''
      self.quotaLimit          = quotaLimit
      self.TIMESTAMPSTART      = TIMESTAMPSTART
      self.TIMESTAMPFINISH     = TIMESTAMPFINISH
      self.TIMERESULTS         = TIMERESULTS


   def __makeSubmitTemplate(self):
      return """#!/bin/sh
#PBS -S /bin/sh
#PBS -l nodes=NNODES:ppn=PPN
#PBS -l walltime=WALLTIME
#PBS -N JOB_JOBID
#PBS -o pbs_JOBID.stdout
#PBS -e pbs_JOBID.stderr
#PBS -mn
REMOTEBATCHACCOUNT
QUEUE

ENVIRONMENT

trap cleanup HUP INT QUIT ABRT TERM

cleanup()
{
   echo "Abnormal termination by signal"
   if [ ! -s TS_FINISH ] ; then
      date +"%s" > TS_FINISH
   fi
   exit 1
}

export PATH=${PBS_O_WORKDIR}:${PBS_O_PATH}
# Change to directory where job was submitted.
cd ${PBS_O_WORKDIR}

TIMEPATH=
for timePath in ${HOME}/bin/time /usr/bin/time /usr/local/bin/time /apps/submit/bin/time ; do
   if [ -x ${timePath} ] ; then
      TIMEPATH=${timePath}
      break
   fi
done

QUOTA

date +"%s" > TS_START

${TIMEPATH} -p -o TIME_RESULTS EXECUTABLE ARGUMENTS STDIN > JOBID.stdout 2> JOBID.stderr

date +"%s" > TS_FINISH
"""


   def __buildSubmitFile(self):
      # setup regex's for the template
      re_stdinput    = re.compile("STDIN")
      re_wallTime    = re.compile("WALLTIME")
      re_queue       = re.compile("QUEUE")
      re_environment = re.compile("ENVIRONMENT")
      re_executable  = re.compile("EXECUTABLE")
      re_arguments   = re.compile("ARGUMENTS")
      re_tsStart     = re.compile("TS_START")
      re_tsFinish    = re.compile("TS_FINISH")
      re_timeResults = re.compile("TIME_RESULTS")
      re_jobid       = re.compile("JOBID")
      re_nnodes      = re.compile("NNODES")
      re_ppn         = re.compile("PPN")
      re_quota       = re.compile("QUOTA")
      re_account     = re.compile("REMOTEBATCHACCOUNT")

      template = self.__makeSubmitTemplate()

      queuePBS = ""
      if self.remoteBatchQueue != "":
         queuePBS = "#PBS -q " + self.remoteBatchQueue

      environmentExport = ""
      environmentVars = self.environment.split()
      for environmentVar in environmentVars:
         environmentExport += "export " + environmentVar + "\n"

      if self.stdinput == "":
          template = re_stdinput.sub(self.stdinput,template)
      else:
          template = re_stdinput.sub(" < "+self.stdinput,template)
      template = re_wallTime.sub(self.wallTime,template)
      template = re_queue.sub(queuePBS,template)
      template = re_environment.sub(environmentExport,template)
      template = re_executable.sub(self.executable,template)
      template = re_arguments.sub(self.arguments,template)
      template = re_tsStart.sub(self.TIMESTAMPSTART,template)
      template = re_tsFinish.sub(self.TIMESTAMPFINISH,template)
      template = re_timeResults.sub(self.TIMERESULTS,template)
      template = re_jobid.sub(self.localJobId,template)
      template = re_nnodes.sub(self.nNodes,template)
      template = re_ppn.sub(self.ppn,template)
      if self.quotaLimit > 0:
         template = re_quota.sub("ulimit -f " + str(self.quotaLimit),template)
      else:
         template = re_quota.sub("",template)
      template = re_account.sub(self.account,template)

      return(template)


   def __makeSubmitMatlabMPITemplate(self):
      return """#!/bin/sh
#PBS -S /bin/sh
#PBS -l nodes=NNODES:ATTRIBUTES:ppn=PPN
#PBS -l walltime=WALLTIME
#PBS -N JOB_JOBID
#PBS -o pbs_JOBID.stdout
#PBS -e pbs_JOBID.stderr
#PBS -mn
REMOTEBATCHACCOUNT
QUEUE

export MCR_INHIBIT_CTF_LOCK=1
ENVIRONMENT

trap cleanup HUP INT QUIT ABRT TERM

cleanup()
{
   echo "Abnormal termination by signal"
   for p in $(ls -r MatMPI/pid.master.*) ; do
      pid=${p:18}
      kill -s TERM $pid
   done
   if [ ! -s TS_FINISH ] ; then
      date +"%s" > TS_FINISH
   fi
   exit 1
}

export PATH=${PBS_O_WORKDIR}:${PBS_O_PATH}
# Change to directory where job was submitted.
cd ${PBS_O_WORKDIR}

TIMEPATH=
for timePath in ${HOME}/bin/time /usr/bin/time /usr/local/bin/time /apps/submit/bin/time ; do
   if [ -x ${timePath} ] ; then
      TIMEPATH=${timePath}
      break
   fi
done

export TIMERCOMMAND="${TIMEPATH} -p -a -o TIME_RESULTS"

date +"%s" > TS_START

PREMANAGERCOMMANDS
MANAGERCOMMAND EXECUTABLE ARGUMENTS
POSTMANAGERCOMMANDS

date +"%s" > TS_FINISH
"""


   def __buildSubmitMatlabMPIFile(self):
      # setup regex's for the template
      re_preManagerCommands  = re.compile("PREMANAGERCOMMANDS")
      re_managerCommand      = re.compile("MANAGERCOMMAND")
      re_postManagerCommands = re.compile("POSTMANAGERCOMMANDS")
      re_wallTime            = re.compile("WALLTIME")
      re_nnodes              = re.compile("NNODES")
      re_attributes          = re.compile(":ATTRIBUTES")
      re_ppn                 = re.compile("PPN")
      re_processors          = re.compile("NPROCESSORS")
      re_queue               = re.compile("QUEUE")
      re_environment         = re.compile("ENVIRONMENT")
      re_tsStart             = re.compile("TS_START")
      re_tsFinish            = re.compile("TS_FINISH")
      re_jobid               = re.compile("JOBID")
      re_executable          = re.compile("EXECUTABLE")
      re_arguments           = re.compile("ARGUMENTS")
      re_timeResults         = re.compile("TIME_RESULTS")
      re_account             = re.compile("REMOTEBATCHACCOUNT")

      template = self.__makeSubmitMatlabMPITemplate()

      queuePBS = ""
      if self.remoteBatchQueue != "":
         queuePBS = "#PBS -q " + self.remoteBatchQueue

      environmentExport = ""
      environmentVars = self.environment.split()
      for environmentVar in environmentVars:
         environmentExport += "export " + environmentVar + "\n"

      template = re_preManagerCommands.sub("\n".join(self.preManagerCommands),template)
      template = re_postManagerCommands.sub("\n".join(self.postManagerCommands),template)
      template = re_managerCommand.sub(self.managerCommand,template)
      template = re_wallTime.sub(self.wallTime,template)
      template = re_nnodes.sub(self.nNodes,template)
      template = re_attributes.sub(self.hostAttributes,template)
      template = re_ppn.sub(self.ppn,template)
      nProcessors = str(int(self.nNodes)*int(self.ppn))
      template = re_processors.sub(nProcessors,template)
      template = re_queue.sub(queuePBS,template)
      template = re_environment.sub(environmentExport,template)
      template = re_executable.sub(self.executable,template)
      template = re_arguments.sub(self.arguments,template)
      template = re_tsStart.sub(self.TIMESTAMPSTART,template)
      template = re_tsFinish.sub(self.TIMESTAMPFINISH,template)
      template = re_jobid.sub(self.localJobId,template)
      template = re_timeResults.sub(self.TIMERESULTS,template)
      template = re_account.sub(self.account,template)

      return(template)


   def __makeSubmitMPITemplate(self):
      return """#!/bin/sh
#PBS -S /bin/sh
#PBS -l nodes=NNODES:ATTRIBUTES:ppn=PPN
#PBS -l walltime=WALLTIME
#PBS -N JOB_JOBID
#PBS -o pbs_JOBID.stdout
#PBS -e pbs_JOBID.stderr
#PBS -mn
REMOTEBATCHACCOUNT
QUEUE

ENVIRONMENT

trap cleanup HUP INT QUIT ABRT TERM

cleanup()
{
   echo "Abnormal termination by signal"
   if [ ! -s TS_FINISH ] ; then
      date +"%s" > TS_FINISH
   fi
   exit 1
}

export PATH=${PBS_O_WORKDIR}:${PBS_O_PATH}
# Change to directory where job was submitted.
cd ${PBS_O_WORKDIR}

chmod +x APPSCRIPTNAME
date +"%s" > TS_START

PREMANAGERCOMMANDS
MANAGERCOMMAND `pwd`/APPSCRIPTNAME > JOBID.stdout 2> JOBID.stderr
POSTMANAGERCOMMANDS

date +"%s" > TS_FINISH

touch TIME_RESULTS-0
cat TIME_RESULTS-[0-9]* >> TIME_RESULTS
rm -f TIME_RESULTS-[0-9]*
"""


   def __buildSubmitMPIFile(self):
      # setup regex's for the template
      re_preManagerCommands  = re.compile("PREMANAGERCOMMANDS")
      re_managerCommand      = re.compile("MANAGERCOMMAND")
      re_postManagerCommands = re.compile("POSTMANAGERCOMMANDS")
      re_wallTime            = re.compile("WALLTIME")
      re_nnodes              = re.compile("NNODES")
      re_attributes          = re.compile(":ATTRIBUTES")
      re_ppn                 = re.compile("PPN")
      re_processors          = re.compile("NPROCESSORS")
      re_queue               = re.compile("QUEUE")
      re_environment         = re.compile("ENVIRONMENT")
      re_appScriptName       = re.compile("APPSCRIPTNAME")
      re_tsStart             = re.compile("TS_START")
      re_tsFinish            = re.compile("TS_FINISH")
      re_timeResults         = re.compile("TIME_RESULTS")
      re_jobid               = re.compile("JOBID")
      re_pbsJobidHide        = re.compile("PBS_JOBID")
      re_pbsJobidShow        = re.compile("PBS_JXOBID")
      re_account             = re.compile("REMOTEBATCHACCOUNT")

      template = self.__makeSubmitMPITemplate()

      queuePBS = ""
      if self.remoteBatchQueue != "":
         queuePBS = "#PBS -q " + self.remoteBatchQueue

      environmentExport = ""
      environmentVars = self.environment.split()
      for environmentVar in environmentVars:
         environmentExport += "export " + environmentVar + "\n"

      template = re_preManagerCommands.sub("\n".join(self.preManagerCommands),template)
      template = re_postManagerCommands.sub("\n".join(self.postManagerCommands),template)
      template = re_managerCommand.sub(self.managerCommand,template)
      template = re_wallTime.sub(self.wallTime,template)
      template = re_nnodes.sub(self.nNodes,template)
      template = re_attributes.sub(self.hostAttributes,template)
      template = re_ppn.sub(self.ppn,template)
      nProcessors = str(int(self.nNodes)*int(self.ppn))
      template = re_processors.sub(nProcessors,template)
      template = re_queue.sub(queuePBS,template)
      template = re_environment.sub(environmentExport,template)
      template = re_appScriptName.sub(self.appScriptName,template)
      template = re_tsStart.sub(self.TIMESTAMPSTART,template)
      template = re_tsFinish.sub(self.TIMESTAMPFINISH,template)
      template = re_timeResults.sub(self.TIMERESULTS,template)
      template = re_pbsJobidHide.sub("PBS_JXOBID",template)
      template = re_jobid.sub(self.localJobId,template)
      template = re_pbsJobidShow.sub("PBS_JOBID",template)
      template = re_account.sub(self.account,template)

      return(template)


   def buildBatchScript(self):
      batchScriptName = self.localJobId + '.pbs'
      if self.isMultiCoreRequest:
         if   self.computationMode == 'mpi':
            batchScript = self.__buildSubmitMPIFile()
         elif self.computationMode == 'parallel':
            batchScript = self.__buildSubmitFile()
         elif self.computationMode == 'matlabmpi':
            batchScript = self.__buildSubmitMatlabMPIFile()
      else:
         batchScript = self.__buildSubmitFile()

      return(batchScriptName,batchScript)


